/*******************************************************************************
 * Copyright (c) 2010 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.report.data.oda.csv.ui.wizards;

import java.io.File;

import org.eclipse.datatools.connectivity.oda.IConnection;
import org.eclipse.datatools.connectivity.oda.IDriver;
import org.eclipse.datatools.connectivity.oda.IQuery;
import org.eclipse.datatools.connectivity.oda.IResultSetMetaData;
import org.eclipse.datatools.connectivity.oda.OdaException;
import org.eclipse.datatools.connectivity.oda.design.DataSetDesign;
import org.eclipse.datatools.connectivity.oda.design.DesignFactory;
import org.eclipse.datatools.connectivity.oda.design.Properties;
import org.eclipse.datatools.connectivity.oda.design.ResultSetColumns;
import org.eclipse.datatools.connectivity.oda.design.ResultSetDefinition;
import org.eclipse.datatools.connectivity.oda.design.ui.designsession.DesignSessionUtil;
import org.eclipse.datatools.connectivity.oda.design.ui.wizards.DataSetWizardPage;
import org.eclipse.birt.report.data.oda.csv.impl.CommonConstants;
import org.eclipse.birt.report.data.oda.csv.impl.Driver;
import org.eclipse.birt.report.data.oda.csv.ui.i18n.Messages;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.FormAttachment;
import org.eclipse.swt.layout.FormData;
import org.eclipse.swt.layout.FormLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Text;

/**
 * Extends the ODA design ui framework to provide a driver-specific
 * custom editor page to create or edit an ODA data set design instance.
 */
public class FileSelectionWizardPage extends DataSetWizardPage 
{
    private static String DEFAULT_MESSAGE = Messages
                .getString( "wizard.defaultMessage.selectFile" ); //$NON-NLS-1$
    private static final String CSV_SUFFIX = ".csv";  //$NON-NLS-1$
    
    private final int DEFAULT_WIDTH = 200;
    private final int DEFAULT_HEIGHT = 200;

    private transient Text fileName = null;
   
    private transient List m_availableList = null;

    /** store latest selected file */
    private File m_selectedFile;

    /**
     * @param pageName
     */
    public FileSelectionWizardPage( String pageName )
    {
        super( pageName );
        setTitle( pageName );
        setMessage( DEFAULT_MESSAGE );

        setPageComplete( false );
    }

    /**
     * @param pageName
     * @param title
     * @param titleImage
     */
    public FileSelectionWizardPage( String pageName, String title,
            ImageDescriptor titleImage )
    {
        super( pageName, title, titleImage );
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.eclipse.datatools.connectivity.oda.design.ui.wizards.DataSetWizardPage#createPageCustomControl(org.eclipse.swt.widgets.Composite)
     */
    public void createPageCustomControl( Composite parent )
    {
        setControl( createPageControl( parent ) );
        initializeControl();
    }

    private void initializeControl()
    {
        /* 
         * Optionally restores the state of a previous design session.
         * Obtains designer state, using
         *      getInitializationDesignerState(); 
         */

        DataSetDesign dataSetDesign = getInitializationDesign();
        if( dataSetDesign == null )
            return; // nothing to initialize

        String queryText = dataSetDesign.getQueryText();
        if( queryText == null )
            return; // nothing to initialize

        updateValuesFromQuery( queryText );
        
        /*
         * Optionally honor the request for an editable or
         * read-only design session
         *      isSessionEditable();
         */
    }

    /*
     * (non-Javadoc)
     * @see org.eclipse.datatools.connectivity.oda.design.ui.wizards.DataSetWizardPage#collectDataSetDesign(org.eclipse.datatools.connectivity.oda.design.DataSetDesign)
     */
    protected DataSetDesign collectDataSetDesign( DataSetDesign design )
    {
        if( ! hasValidData() )
            return design;
        savePage( design );
        return design;
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.datatools.connectivity.oda.design.internal.ui.DataSetWizardPage#collectResponseState()
     */
    protected void collectResponseState()
    {        
        super.collectResponseState();
        /* 
         * Optionally assigns custom response state, for inclusion
         * in the ODA design session response, using
         *      setResponseSessionStatus( SessionStatus status )
         *      setResponseDesignerState( DesignerState customState ); 
         */
    }

    /* (non-Javadoc)
     * @see org.eclipse.datatools.connectivity.oda.design.ui.wizards.DataSetWizardPage#canLeave()
     */
    protected boolean canLeave()
    {
        if( ! isPageComplete() )
        {
            setMessage( Messages.getString( "error.selectColumns" ), ERROR ); //$NON-NLS-1$
        }
        return isPageComplete();
    }

    private Control createPageControl( Composite parent )
    {
    	
        Composite composite = new Composite(parent, SWT.NULL);
        
        FormLayout layout = new FormLayout();
        composite.setLayout(layout);
        
        FormData data = new FormData();
        data.left = new FormAttachment(0, 5);
        data.top = new FormAttachment(0, 5);
        
        fileName = new Text(composite, SWT.BORDER);
        fileName.setLayoutData(data);
        Properties dataSourceProps = getInitializationDesign().getDataSourceDesign().getPublicProperties();
        fileName.setText((String)(dataSourceProps.getProperty(Constants.ODAHOME)));
        


        data = new FormData();
        data.top = new FormAttachment(fileName, 10, SWT.BOTTOM);
        data.left = new FormAttachment(0, 5);
        data.right = new FormAttachment(47, -5);
        data.bottom = new FormAttachment(100, -5);
        data.width = DEFAULT_WIDTH;
        data.height = DEFAULT_HEIGHT;
        m_availableList = new List( composite, SWT.MULTI
    			| SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL );

        m_availableList.setLayoutData(data);        
        m_selectedFile = new File((String)(dataSourceProps.getProperty(Constants.ODAHOME)));   
        populateAvailableList();
        
        return composite;    	
    }

    private void populateAvailableList()
    {

    	File file = m_selectedFile;
    	setPageComplete( true );

    	String[] columnNames = getFileColumnNames(file);
    	m_availableList.removeAll();
    	if(columnNames == null || columnNames.length == 0)
    	{
    		setMessage(Messages.getString("error.noColumns"), ERROR); //$NON-NLS-1$
    		disableAll();
    	}
    	else
    	{
    		enableAll();
    		m_availableList.setItems(columnNames);
    		m_availableList.select(0);
    		setMessage(DEFAULT_MESSAGE);
    	}
    }    

    /**
     * Returns all the column names found in given flatfile.
     * @param file
     * @return
     */
    private String[] getFileColumnNames( File file )
    {
        return getQueryColumnNames( "select * from " + file.getName(), file ); //$NON-NLS-1$
    }

    /**
     * Returns the column names selected by the given query
     * accessing the given flatfile.
     */
    private String[] getQueryColumnNames( String queryText, File file )
    {
        IDriver ffDriver = new Driver();
        IConnection conn = null;
        try
        {
            conn = ffDriver.getConnection( null );
            IResultSetMetaData metadata = getResultSetMetaData( queryText, file, conn );
            int columnCount = metadata.getColumnCount();
            if( columnCount == 0 )
                return null;

            String[] result = new String[columnCount];
            for( int i = 0; i < columnCount; i++)
                result[i] = metadata.getColumnName( i + 1 );
            return result;
        }
        catch( OdaException e )
        {
        	setMessage( e.getLocalizedMessage( ), ERROR ); //$NON-NLS-1$
            disableAll();
            return null;
        }
        finally
        {
            closeConnection( conn );
        }
    }

    /**
     * Attempts to close given ODA connection.
     * @param conn
     */
    private void closeConnection( IConnection conn )
    {
        try
        {
            if( conn != null )
                conn.close();
        }
        catch( OdaException e )
        {
            // ignore
        }
    }

    /**
     * Obtains the query's result set metadata from
     * the ODA flatfile runtime driver by preparing
     * the given query.
     * Uses the data source connection properties setting.
     * @throws OdaException
     */
    private IResultSetMetaData getResultSetMetaData( 
                String queryText,
                File file, IConnection conn ) throws OdaException
    {
        java.util.Properties prop = new java.util.Properties();
        prop.put( CommonConstants.CONN_HOME_DIR_PROP, file.getParent() );
        conn.open( prop );

        IQuery query = conn.newQuery( null );
        query.setMaxRows( Constants.DEFAULT_MAX_ROWS );
        query.prepare( queryText );
        query.executeQuery();

        return query.getMetaData();
    }

    /**
     * Enable all control of this page
     */
    private void enableAll()
    {
        m_availableList.setEnabled( true );
    }

    /**
     * Disable all control of this page
     */
    private void disableAll()
    {
        m_availableList.setEnabled( false );
        setPageComplete( false );
    }

 
    /**
     * Return query of this data set
     * 
     * @return
     */
    private String getQuery()
    {
        String tableName = null;
        StringBuffer buf = new StringBuffer();
        File file = m_selectedFile; 
        
        if(file != null)
        {
            tableName = file.getName();
        }
        if(tableName != null)
        {
            if(m_availableList.getItemCount() == 0)
            {
                buf.append("select * from ").append(tableName); //$NON-NLS-1$
            }
            else
            {
                buf.append("select "); //$NON-NLS-1$
                String[] columns = m_availableList.getItems();
                for(int n = 0; n < columns.length; n++)
                {
                    buf.append(columns[n]);
                    if(n < columns.length - 1)
                    {
                        buf.append(", "); //$NON-NLS-1$
                    }
                }
                buf.append(" from ").append(tableName); //$NON-NLS-1$
            }
        }
        return buf.toString();
    	
 
    }

    /**
     * Update value of query text
     * 
     * @param query
     */
    private void updateValuesFromQuery( String query )
    {
        String[] splitQuery = stripFromClause(query);
        //If the length is equal to 2 then we have a valid query
        if(splitQuery.length == 2)
        {
            String table = splitQuery[1];
            String columns = stripSelect(splitQuery[0]);
            //If both are not null then we can assume we have a valid query
            if(table != null && columns != null)
            {
                //Now we need to select the table in the list
                //If it doesn't exists then we do not need to process the columns
                File f = selectTableFromQuery(table); 
                if(f != null)
                {
                    updateColumnsFromQuery(getColumnsFromQuery(columns), getFileColumnNames(f));
                }
            }
        }
        
        if(m_availableList.getItemCount() == 0)
        {
            setPageComplete(false);
        }        
    }

    /**
     * Split Select clause of query text
     * 
     * @param selectedColumns
     * @return
     */
    private String stripSelect( String selectedColumns )
    {
        String[] result = selectedColumns.split( "SELECT " ); //$NON-NLS-1$
        if( result.length == 2 )
        {
            return result[1];
        }

        return null;
    }

    /**
     * Split From clause of query text
     * 
     * @param selectedColumns
     * @return
     */
    private static String[] stripFromClause( String query )
    {
        query = query.toUpperCase();
        return query.split( " FROM " ); //$NON-NLS-1$
    }

    /**
     * @param tableName
     * @return File
     */
    private File selectTableFromQuery( String tableName )
    {
        Properties dataSourceProps = getInitializationDesign().getDataSourceDesign().getPublicProperties();
        File file = new File((String)(dataSourceProps.getProperty(Constants.ODAHOME)));   
        return file;
    }

    /**
     * Update selected and unselected column from lastest query text
     * 
     * @param columns
     * @param availableColumns
     */
    private void updateColumnsFromQuery( String[] columns,
            String[] availableColumns )
    {
        m_availableList.setItems(availableColumns);
        for(int n = 0; n < columns.length; n++)
        {
            if(columns[n].trim().equals("*")) //$NON-NLS-1$
            {
                //WildCard just select everything
                setPageComplete(true);
                 return;
            }
        }
        setPageComplete(true);   	
    }

    /**
     * Return all columns from column list
     * @param columnList
     * @return
     */
    private String[] getColumnsFromQuery( String columnList )
    {
        String[] columns = columnList.split( CommonConstants.DELIMITER_COMMA ); //$NON-NLS-1$
        // Remove the as keyword if any
        for( int n = 0; n < columns.length; n++)
        {
            String[] result = columns[n].split( " AS " ); //$NON-NLS-1$
            columns[n] = result[0].trim();
        }

        return columns;
    }


    private boolean hasValidData()
    {
        Properties dataSourceProps = getInitializationDesign().getDataSourceDesign().getPublicProperties();
        File file = new File((String)(dataSourceProps.getProperty(Constants.ODAHOME)));   

        String fileName = file.getName().toLowerCase();
        if( !fileName.endsWith( CSV_SUFFIX ) )  
        {
            if( !isContinue( Messages
                    .getString( "warning.columnExtensionInvalid" ) ) ) //$NON-NLS-1$
                return false;
        }
        if( isPageComplete() )
        {
            return true;
        }
        setMessage( Messages.getString( "error.selectColumns" ), ERROR ); //$NON-NLS-1$
        return false;
    }

 
    /**
     * Warn user whether action continues
     * 
     * @param warnMsg
     * @return true, continue action
     */
    private boolean isContinue( String warnMsg )
    {
        MessageDialog dialog = new MessageDialog( this.getShell(), Messages
                .getString( "title.warning" ), null, warnMsg, //$NON-NLS-1$
                MessageDialog.WARNING, new String[]
                { IDialogConstants.YES_LABEL, IDialogConstants.NO_LABEL }, 0 );
        return dialog.open() == MessageDialog.OK;
    }

    /**
     * Updates the given dataSetDesign with the query and its metadata defined
     * in this page.
     * @param dataSetDesign
     */
    private void savePage( DataSetDesign dataSetDesign )
    {
        String queryText = getQuery();
        dataSetDesign.setQueryText( queryText );

        // obtain query's result set metadata, and update
        // the dataSetDesign with it
        IConnection conn = null;
        try
        {
            IDriver ffDriver = new Driver();
            conn = ffDriver.getConnection( null );
            IResultSetMetaData metadata = 
                getResultSetMetaData( queryText, m_selectedFile, conn );
            setResultSetMetaData( dataSetDesign, metadata );
        }
        catch( OdaException e )
        {
            // no result set definition available, reset in dataSetDesign
            dataSetDesign.setResultSets( null );
        }
        finally
        {
            closeConnection( conn );
        }
        
         /*
         * Since this csv file driver does not support
         * query parameters and properties, there are
         * no data set parameters and public/private properties
         * to specify in the data set design instance
         */
    }

    private void setResultSetMetaData( DataSetDesign dataSetDesign,
            IResultSetMetaData md ) throws OdaException
    {
        ResultSetColumns columns = DesignSessionUtil.toResultSetColumnsDesign( md );

        ResultSetDefinition resultSetDefn = DesignFactory.eINSTANCE
                .createResultSetDefinition();
        // csv file does not support result set name
        resultSetDefn.setResultSetColumns( columns );

        // no exception; go ahead and assign to specified dataSetDesign
        dataSetDesign.setPrimaryResultSet( resultSetDefn );
        dataSetDesign.getResultSets().setDerivedMetaData( true );
    }
}
