/*******************************************************************************
 * Copyright (c) 2010 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.report.data.oda.csv.impl;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.math.BigDecimal;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Vector;

import org.eclipse.birt.report.data.oda.csv.i18n.Messages;
import org.eclipse.datatools.connectivity.oda.IConnection;
import org.eclipse.datatools.connectivity.oda.IParameterMetaData;
import org.eclipse.datatools.connectivity.oda.IResultSet;
import org.eclipse.datatools.connectivity.oda.IResultSetMetaData;
import org.eclipse.datatools.connectivity.oda.IQuery;
import org.eclipse.datatools.connectivity.oda.OdaException;
import org.eclipse.datatools.connectivity.oda.SortSpec;
import org.eclipse.datatools.connectivity.oda.spec.QuerySpecification;

/**
 * Implementation class of IQuery for an ODA runtime driver.
 * <br>
 * For demo purpose, the auto-generated method stubs have
 * hard-coded implementation that returns a pre-defined set
 * of meta-data and query results.
 * A custom ODA driver is expected to implement own data source specific
 * behavior in its place. 
 */
public class Query implements IQuery
{
	public static final int DEFAULT_MAX_ROWS = 1000;
	
	private static final String NAME_LITERAL = "NAME"; //$NON-NLS-1$
	private static final String TYPE_LITERAL = "TYPE"; //$NON-NLS-1$
	private static final char QUERY_TEXT_DELIMITER = ':';
	private static final char COLUMNSINFO_BEGIN_DELIMITER = '{';

	private int m_maxRows;
	// The home directory specified in associated connection
	private String homeDirectory = null;
	// The table that the query operates on
	private String currentTableName = null;
	// The Connection instance associated with the Query.
	private IConnection connection = null;
	// The meta data of the query's result set.
	// It is available only after a query is prepared.
	private IResultSetMetaData resultSetMetaData = null;
	
	private String charSet = CommonConstants.CONN_DEFAULT_CHARSET;

	public Query( String homeDir, IConnection host ) throws OdaException
	{
		if ( homeDir == null || host == null )
			throw new OdaException( Messages.getString( "common_ARGUMENT_CANNOT_BE_NULL" ) ); //$NON-NLS-1$
		File file = new File(homeDir);
		if (file.isDirectory())		
			this.homeDirectory = homeDir;
		else if (file.isFile())	
			this.homeDirectory = file.getParent();
		this.connection = host;

	}
	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#prepare(java.lang.String)
	 */
	public void prepare( String queryText ) throws OdaException
	{
		if ( queryText != null )
		{
			String query = splitQueryText(queryText)[0] ;
			String colInfo = splitQueryText( queryText )[1];
			validateOpenConnection( );
			String formattedQuery = formatQueryText( query );
			validateQueryText( formattedQuery );
			prepareMetaData( formattedQuery, colInfo );
		}
		else
			throw new OdaException( Messages.getString( "common_NULL_QUERY_TEXT" ) );
	}
	
	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setAppContext(java.lang.Object)
	 */
	public void setAppContext( Object context ) throws OdaException
	{
	    // do nothing; assumes no support for pass-through context
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#close()
	 */
	public void close() throws OdaException
	{
		currentTableName = null;
		m_maxRows = 0;
		connection = null;
		resultSetMetaData = null;
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#getMetaData()
	 */
	public IResultSetMetaData getMetaData() throws OdaException
	{
        /* TODO Auto-generated method stub
         * Replace with implementation to return an instance 
         * based on this prepared query.
         */
		return this.resultSetMetaData;
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#executeQuery()
	 */
	public IResultSet executeQuery() throws OdaException
	{
 		
		// fetch data from file to a Vector
		Vector v = fetchQueriedDataFromFileToVector( );

		// initialize a String array with data from a Vector
		String[][] rowSet = copyDataFromVectorToTwoDimensionArray( v );

		return new ResultSet( rowSet, this.resultSetMetaData );
		
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setProperty(java.lang.String, java.lang.String)
	 */
	public void setProperty( String name, String value ) throws OdaException
	{
		// do nothing; assumes no data set query property
		throw new UnsupportedOperationException( );

	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setMaxRows(int)
	 */
	public void setMaxRows( int max ) throws OdaException
	{
	    m_maxRows = max;
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#getMaxRows()
	 */
	public int getMaxRows() throws OdaException
	{
		return m_maxRows;
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#clearInParameters()
	 */
	public void clearInParameters() throws OdaException
	{
        // TODO Auto-generated method stub
		// only applies to input parameter
		throw new UnsupportedOperationException( );
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setInt(java.lang.String, int)
	 */
	public void setInt( String parameterName, int value ) throws OdaException
	{
        // TODO Auto-generated method stub
		// only applies to named input parameter
		throw new UnsupportedOperationException( );

	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setInt(int, int)
	 */
	public void setInt( int parameterId, int value ) throws OdaException
	{
        // TODO Auto-generated method stub
		// only applies to input parameter
		throw new UnsupportedOperationException( );

	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setDouble(java.lang.String, double)
	 */
	public void setDouble( String parameterName, double value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to named input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setDouble(int, double)
	 */
	public void setDouble( int parameterId, double value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setBigDecimal(java.lang.String, java.math.BigDecimal)
	 */
	public void setBigDecimal( String parameterName, BigDecimal value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to named input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setBigDecimal(int, java.math.BigDecimal)
	 */
	public void setBigDecimal( int parameterId, BigDecimal value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setString(java.lang.String, java.lang.String)
	 */
	public void setString( String parameterName, String value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to named input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setString(int, java.lang.String)
	 */
	public void setString( int parameterId, String value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setDate(java.lang.String, java.sql.Date)
	 */
	public void setDate( String parameterName, Date value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to named input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setDate(int, java.sql.Date)
	 */
	public void setDate( int parameterId, Date value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setTime(java.lang.String, java.sql.Time)
	 */
	public void setTime( String parameterName, Time value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to named input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setTime(int, java.sql.Time)
	 */
	public void setTime( int parameterId, Time value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setTimestamp(java.lang.String, java.sql.Timestamp)
	 */
	public void setTimestamp( String parameterName, Timestamp value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to named input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setTimestamp(int, java.sql.Timestamp)
	 */
	public void setTimestamp( int parameterId, Timestamp value ) throws OdaException
	{
		throw new UnsupportedOperationException( );

        // TODO Auto-generated method stub
		// only applies to input parameter
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#findInParameter(java.lang.String)
	 */
	public int findInParameter( String parameterName ) throws OdaException
	{
        // TODO Auto-generated method stub
		// only applies to named input parameter
		throw new UnsupportedOperationException( );
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#getParameterMetaData()
	 */
	public IParameterMetaData getParameterMetaData() throws OdaException
	{
        /* TODO Auto-generated method stub
         * Replace with implementation to return an instance 
         * based on this prepared query.
         */
		throw new UnsupportedOperationException( );
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#setSortSpec(org.eclipse.datatools.connectivity.oda.SortSpec)
	 */
	public void setSortSpec( SortSpec sortBy ) throws OdaException
	{
        // TODO Auto-generated method stub
		// only applies to sorting, assumes not supported
        throw new UnsupportedOperationException();
	}

	/*
	 * @see org.eclipse.datatools.connectivity.oda.IQuery#getSortSpec()
	 */
	public SortSpec getSortSpec() throws OdaException
	{
        // TODO Auto-generated method stub
		// only applies to sorting
		throw new UnsupportedOperationException( );
	}
	
	
	/**
	 * Validate whether the query's connection is open.
	 * 
	 * @throws OdaException
	 *             if connection is not open yet
	 */
	private void validateOpenConnection( ) throws OdaException
	{
		if ( connection.isOpen( ) == false )
			throw new OdaException( Messages.getString( "common_CONNECTION_HAS_NOT_OPENED" ) ); //$NON-NLS-1$
	}
	/**
	 * Format the given query text. Eliminates redundant spaces and convert all
	 * keywords to uppercase.
	 * 
	 * @param queryText
	 * @return
	 */
	private String formatQueryText( String queryText )
	{
		String result = ""; //$NON-NLS-1$
		String[] temp = queryText.trim( )
				.split( CommonConstants.DELIMITER_SPACE );
		for ( int i = 0; i < temp.length; i++ )
		{
			if ( temp[i].equalsIgnoreCase( CommonConstants.KEYWORD_AS ) )
				temp[i] = temp[i].toUpperCase( );
			if ( temp[i].equalsIgnoreCase( CommonConstants.KEYWORD_FROM ) )
				temp[i] = temp[i].toUpperCase( );
			if ( temp[i].equalsIgnoreCase( CommonConstants.KEYWORD_SELECT ) )
				temp[i] = temp[i].toUpperCase( );
			result = result + temp[i] + CommonConstants.DELIMITER_SPACE;
		}
		return result.trim( );
	}
	/**
	 * Validate whether the query text is valid.
	 * 
	 * @param formattedQuery
	 *            the SQL command
	 * @throws OdaException
	 *             if given text is invalid or if any other error occurs
	 */
	private void validateQueryText( String formattedQuery ) throws OdaException
	{
		validateNonEmptyQueryText( formattedQuery );

		String[] queryFragments = parsePreparedQueryText( formattedQuery );

		if ( queryFragments[2].split( CommonConstants.DELIMITER_COMMA ).length != 1 )
			throw new OdaException( Messages.getString( "query_DO_NOT_SUPPORT_CROSS_TABLE_QUERY" ) );

	}
	/**
	 * Validate whether the given query text is empty.
	 * 
	 * @param formattedQuery
	 *            the trimed query text
	 * @throws OdaException
	 *             if the given text is empty
	 */
	private void validateNonEmptyQueryText( String formattedQuery )
			throws OdaException
	{
		if ( formattedQuery == null || formattedQuery.length( ) == 0 )
			throw new OdaException( Messages.getString( "query_COMMAND_IS_EMPTY" ) ); //$NON-NLS-1$
	}
	/**
	 * Parse the command by separating keywords and other parts of a SQL SELECT
	 * query text.
	 * 
	 * @param formattedQuery
	 *            SQL SELECT query: SELECT COLUMNNAME (AS ALIAS)[,COLUMNNAME2
	 *            (AS ALIAS)] FROM TABLENAME
	 * @return a String array with first element that holds all the
	 *         comma-separated column names, the second element that holds all
	 *         the comma-separated column display labels, and the third element
	 *         for table name(s) in FROM clause
	 * @throws OdaException
	 *             if the given query is not valid.
	 */
	private String[] parsePreparedQueryText( String formattedQuery )
			throws OdaException
	{
		// This array stores two values: "SELECT" keyword and other part of a
		// command
		String[] queryWithoutSELECTKeyword = formattedQuery.split( CommonConstants.DELIMITER_SPACE, 2 );
		if ( queryWithoutSELECTKeyword == null
				|| queryWithoutSELECTKeyword.length != 2
				|| !queryWithoutSELECTKeyword[0].trim( )
						.equalsIgnoreCase( CommonConstants.KEYWORD_SELECT ) )
			throw new OdaException( Messages.getString( "query_COMMAND_NOT_VALID" ) ); //$NON-NLS-1$

		/*
		 * This returned array stores two values: the fragment immediately after
		 * "SELECT" and before "FROM", and the fragment immediate after the
		 * "FROM" keyword.
		 */
		String[] querySelectAndFromFragments = queryWithoutSELECTKeyword[1].split( CommonConstants.DELIMITER_SPACE
				+ CommonConstants.KEYWORD_FROM
				+ CommonConstants.DELIMITER_SPACE );
		if ( querySelectAndFromFragments == null || querySelectAndFromFragments.length != 2 )
			throw new OdaException( Messages.getString( "query_COMMAND_NOT_VALID" ) ); //$NON-NLS-1$
		
		String[] result = new String[3];
		// store the table name in given last element as the third element
		result[2] = querySelectAndFromFragments[1];

		// split the columns specified in the SELECT clause
		String selectedColumns = querySelectAndFromFragments[0];
		if ( !isWildCard( selectedColumns ) )
		{
			String[] columns = selectedColumns.split( CommonConstants.DELIMITER_COMMA );
			for ( int i = 0; i < columns.length; i++ )
			{
				String[] columnNameAlias = columns[i].split( CommonConstants.DELIMITER_SPACE
						+ CommonConstants.KEYWORD_AS
						+ CommonConstants.DELIMITER_SPACE );
				if ( columnNameAlias != null )
				{
					// append column name to comma-separated column names in
					// result[0]
					result[0] = ( i == 0 ? columnNameAlias[0] : result[0]
							+ CommonConstants.DELIMITER_COMMA
							+ columnNameAlias[0].trim( ) );

					// append column alias, if exists, or null to
					// comma-separated column aliases in result[1]
					if ( columnNameAlias.length == 2 )
						result[1] = ( i == 0 ? columnNameAlias[1] : result[1]
								+ CommonConstants.DELIMITER_COMMA
								+ columnNameAlias[1].trim( ) );
					else
						result[1] = ( i == 0 ? null : result[1]
								+ CommonConstants.DELIMITER_COMMA + null );
				}
			}
		}
		else
		{
			result[0] = CommonConstants.KEYWORD_ASTERISK;
			result[1] = null;
		}

		return result;

	}

	/**
	 * @param cCN
	 * @return
	 */
	private boolean isWildCard( String cCN )
	{
		if ( cCN.equalsIgnoreCase( CommonConstants.KEYWORD_ASTERISK ) )
			return true;
		return false;
	}
	
	/**
	 * Feed the row data from a Vector to a two-dimension array. The string
	 * value is trimmed before being copied into array.
	 * 
	 * @param v
	 * @return a String two dimension array with each horizontal array contains
	 *         a row
	 * @throws OdaException
	 */
	private String[][] copyDataFromVectorToTwoDimensionArray( Vector v )
			throws OdaException
	{
		String[][] rowSet = new String[v.size( )][this.resultSetMetaData.getColumnCount( )];
		for ( int i = 0; i < v.size( ); i++ )
		{
			String[] temp = (String[]) v.elementAt( i );
			for ( int j = 0; j < temp.length; j++ )
			{
				rowSet[i][j] = temp[j].trim( );
			}
		}
		return rowSet;
	}

	
	/**
	 * Prepare the meta data which will be used in execution of a query text. It
	 * sets the value of two member variables: resultSetMetaData and
	 * currentTableName
	 * 
	 * @param queryText
	 * @throws OdaException
	 */
	private void prepareMetaData( String query, String savedSelectedColInfo )
			throws OdaException
	{
		String[] queryFragments = parsePreparedQueryText( query );
		// the name of table against which the query will be executed
		String tableName = queryFragments[2];

		// the array that contains the actual column names read from data file
		String[] allColumnNames = discoverActualColumnMetaData( tableName,
				NAME_LITERAL ); 

		String[] allColumnTypes = createTempColumnTypes( allColumnNames.length );	
		// the array that contains the column names read from command
		String[] queryColumnNames = null;
		String[] queryColumnTypes = null;
		String[] queryColumnLables = null;

		queryColumnNames = allColumnNames;
		queryColumnTypes = allColumnTypes;
		queryColumnLables = allColumnNames;

		this.resultSetMetaData = new ResultSetMetaData( queryColumnNames,
				queryColumnTypes, queryColumnLables );
		this.currentTableName = tableName;

		///*********************

	}
	
	/**
	 * Returns a specified array of metadata info
	 * 
	 * @param tableName
	 * @param metaDataType:
	 *            currently has two values: "NAME" and "TYPE"
	 * @return String[] an array that holds the specified metadata
	 * @throws OdaException
	 */
	private String[] discoverActualColumnMetaData( String tableName,
			String metaDataType ) throws OdaException
	{
		try
		{
			String dataFilePath = findDataFileAbsolutePath( tableName );
			FileInputStream fis = new FileInputStream( dataFilePath );
			InputStreamReader isr = new InputStreamReader( fis, CommonConstants.CONN_DEFAULT_CHARSET );
			CSVBufferedReader br = new CSVBufferedReader( isr );

			if ( !( metaDataType.trim( ).equalsIgnoreCase( NAME_LITERAL ) || metaDataType.trim( ).equalsIgnoreCase( TYPE_LITERAL ) ) )  
				throw new OdaException( Messages.getString( "query_ARGUMENT_ERROR" ) ); //$NON-NLS-1$

			// if want to discover type information then just skip first line
			String[] result = br.readLine( ).split(CommonConstants.DELIMITER_COMMA);

			br.close( );

			if ( metaDataType.trim( ).equalsIgnoreCase( NAME_LITERAL ) ) 
				this.validateUniqueName( result );

			return trimStringArray( result );

		}

		catch ( IOException e )
		{
			throw new OdaException( Messages.getString( "query_IO_EXCEPTION" ) //$NON-NLS-1$
					+ findDataFileAbsolutePath( tableName ) );
		}

	}

	/**
	 * 
	 * @param columnCount
	 * @return
	 */
	private String[] createTempColumnTypes( int columnCount )
	{
		String[] tempColumnTypes = new String[columnCount];

		for ( int i = 0; i < columnCount; i++ )
		{
			tempColumnTypes[i] = "STRING";
		}

		return tempColumnTypes;
	}
	/**
	 * Validate that there are no duplicate column names in a table
	 * 
	 * @param aCN
	 * @throws OdaException
	 */
	private void validateUniqueName( String[] aCN ) throws OdaException
	{
		for ( int i = 0; i < aCN.length; i++ )
		{
			if ( this.findOccuranceOfValueInStringArray( aCN[i], aCN ) > 1 )
			{
				throw new OdaException( Messages.getString( "query_SOURCE_DATA_ERROR" ) ); //$NON-NLS-1$
			}
		}
	}

	/**
	 * Return the number of occurance of a value in the given array
	 * 
	 * @param value
	 * @param array
	 * @return
	 */
	private int findOccuranceOfValueInStringArray( String value, String[] array )
	{
		int count = 0;
		for ( int i = 0; i < array.length; i++ )
		{
			if ( value.trim( ).equalsIgnoreCase( array[i].trim( ) ) )
				count++;
		}
		return count;
	}
	
	/**
	 * Find the absolute path of the file in which a specific table resides.
	 * 
	 * @param tableName
	 *            the name of table
	 * @return the String which contains the absolute path of that table
	 * @throws OdaException
	 *             if the table name cannot be found
	 */
	private String findDataFileAbsolutePath( String tableName )
			throws OdaException
	{
		File file = new File( this.homeDirectory 
				+ File.separator + tableName.trim( ) );
		if ( !file.exists( ) )
			throw new OdaException( Messages.getString("query_invalidTableName") + tableName ); //$NON-NLS-1$
		return file.getAbsolutePath( );
	}

	/**
	 * 
	 */
	private static String[] splitQueryText( String queryText )
			throws OdaException
	{
		int delimiterIndex = -1;
		int columnsInfoBeginIndex = -1;
		
		String trimmedQueryText = queryText.trim( );
		
		String[] splittedQueryText = {
				"", ""
		};
		boolean inQuote = false;
		boolean isEscaped = false;
		char[] chars = trimmedQueryText.toCharArray( );
		
		for ( int i = 0; i < chars.length; i++ )
		{
			if ( chars[i] == '"' )
			{
				if ( !isEscaped )
					inQuote = !inQuote;
				else
					isEscaped = !isEscaped;
			}
			else if ( chars[i] == '\\' )
			{
				isEscaped = !isEscaped;
			}
			else if ( ( !inQuote ) && chars[i] == QUERY_TEXT_DELIMITER )
				delimiterIndex = i;
			else if ( ( !inQuote ) && chars[i] == COLUMNSINFO_BEGIN_DELIMITER )
			{
				columnsInfoBeginIndex = i;
				break;
			}
		}

		if ( inQuote )
			throw new OdaException( Messages.getString( "query_text_error" ) );

		if ( delimiterIndex != -1
				&& columnsInfoBeginIndex != -1 )
		{
			splittedQueryText[0] = trimmedQueryText.substring( 0, delimiterIndex )
					.trim( );
			splittedQueryText[1] = trimmedQueryText.substring( columnsInfoBeginIndex + 1,
					trimmedQueryText.length( )-1 )
					.trim( );
		}
		else if ( delimiterIndex == -1
				&& columnsInfoBeginIndex == -1 )
			splittedQueryText[0] = trimmedQueryText;
		else
			throw new OdaException( Messages.getString( "query_text_error" ) );

		return splittedQueryText;
	}
	/**
	 * Return the 0-based position of a value in the given array
	 * 
	 * @param value
	 * @param array
	 * @return
	 * @throws OdaException
	 */
	private int findLocationOfValueInStringArray( String value, String[] array )
			throws OdaException
	{
		int result = -1;
		try
		{
			for ( int i = 0; i < array.length; i++ )
			{
				if ( value.trim( ).equalsIgnoreCase( array[i].trim( ) ) )
				{
					result = i;
					break;
				}
			}
			return result;
		}
		catch ( Exception e )
		{
			throw new OdaException( Messages.getString( "query_COLUMN_NAME_ERROR" ) ); //$NON-NLS-1$
		}
	}
	/**
	 * Fetch queried data from source file and return in a vector
	 * 
	 * @return a Vector which contains queried data
	 * @throws OdaException
	 */
	private Vector fetchQueriedDataFromFileToVector( ) throws OdaException
	{
		Vector result = new Vector( );

		try
		{
			String dataFilePath = findDataFileAbsolutePath( currentTableName );
			FileInputStream fis = new FileInputStream( dataFilePath );

			InputStreamReader isr = new InputStreamReader( fis, this.charSet );

			CSVBufferedReader br = new CSVBufferedReader( isr );

			// make a copy of column names.
			String[] columnNames =  br.readLine( ).split(CommonConstants.DELIMITER_COMMA);

			// temporary variable which is used to store the data of a row
			// fetched from a csv file
			String aLine = null;

			int fetchCounter = 0;
			while ( ( aLine = br.readLine( ) ) != null
					&& ( this.m_maxRows <= 0 ? true : fetchCounter < this.m_maxRows ) )
			{
				fetchCounter++;
				result.add( fetchQueriedDataFromRow( aLine, columnNames ) );
			}
			br.close( );
			return result;
		}
		catch ( IOException e )
		{
			throw new OdaException( e.getMessage( ) );
		}
	}


	/**
	 * Fetch data from a row.
	 * 
	 * @param aRow
	 *            a row read from table
	 * @param columnNames
	 *            an array that contains all column names of a table
	 * @return an array of data values for each specified column names from a
	 *         row. The "specified column names" are obtained from meta data
	 * @throws OdaException
	 */
	private String[] fetchQueriedDataFromRow( String aRow, String[] columnNames )
			throws OdaException
	{
		String[] sArray = new String[this.getMetaData( ).getColumnCount( )];
		// Ignore all quotes which is used in the input file for the
		// clarity of the data
		Vector vTemp = splitStringWithDoubleQuotes( aRow );

		for ( int i = 0; i < sArray.length; i++ )
		{
			int location = findLocationOfValueInStringArray( getMetaData( ).getColumnName( i + 1 ),
					columnNames );
			if ( location != -1 )
			{
				if( location >= vTemp.size( ) )
					throw new OdaException( Messages.getString( "query_INVALID_FLAT_FILE" ));
				sArray[i] = vTemp.elementAt( location ).toString( );
			}
		}
		return sArray;
	}

	private Vector splitStringWithDoubleQuotes( String aRow )
	{
		String[] aRowArray = aRow.split( CommonConstants.DELIMITER_COMMA, -1 );
		Vector v = new Vector( );
		String temp = null;
		boolean isPartOfString = false;

		for ( int i = 0; i < aRowArray.length; i++ )
		{
			if ( aRowArray[i].trim( )
					.startsWith( CommonConstants.DELIMITER_DOUBLEQUOTE )
					&& ( !isPartOfString ) )
			{
				if ( !aRowArray[i].trim( )
						.endsWith( CommonConstants.DELIMITER_DOUBLEQUOTE ) )
				{
					temp = aRowArray[i];
					isPartOfString = true;
					continue;
				}
				aRowArray[i] = aRowArray[i].trim( ).substring( 1,
						aRowArray[i].trim( ).length( ) - 1 );
			}
			if ( isPartOfString )
			{
				temp = temp + CommonConstants.DELIMITER_COMMA + aRowArray[i];
				if ( aRowArray[i].trim( )
						.endsWith( CommonConstants.DELIMITER_DOUBLEQUOTE ) )
				{
					v.addElement( temp.trim( ).substring( 1,
							temp.trim( ).length( ) - 1 ) );
					temp = null;
					isPartOfString = false;
				}
			}
			else
			{
				// Fields that contain consecutive quote characters is escaped.
				// Replace a pair of consecutive to one single double quote
				aRowArray[i] = aRowArray[i].replaceAll( "\\Q\"\"\\E", "\"" ); //$NON-NLS-1$ //$NON-NLS-2$
				v.addElement( aRowArray[i] );
			}
		}
		return v;
	}

	
	
	
	/**
	 * @param result
	 * @return
	 */
	private String[] trimStringArray( String[] array )
	{
		String[] result = new String[array.length];
		for ( int i = 0; i < result.length; i++ )
			result[i] = array[i].trim( );
		return result;
	}
	/**
	 * CVSBufferReader will buffer the input from the specified file. Read a
	 * line of text. A line is considered to be terminated by any one of a line
	 * feed ('\n'), or a carriage return followed immediately by a linefeed.
	 */
	public static class CSVBufferedReader
	{

		//
		private Reader reader;
		private char[] charBuffer;
		private int startingPosition;
		private int eofInPosition;
		private static int CHARBUFFSIZE = 8192;

		/**
		 * Constructor.
		 * 
		 * @param in
		 */
		public CSVBufferedReader( Reader in )
		{
			reader = new BufferedReader( in );

			startingPosition = -1;

			eofInPosition = -2;
		}

		/**
		 * Read a line from input stream. The line-seperator is '\n'
		 */
		public String readLine( ) throws IOException
		{
			if ( isLastCharBuff( ) && needRefillCharBuff( ) )
				return null;

			if ( needRefillCharBuff( ) )
			{
				charBuffer = newACharBuff( );
				int close = reader.read( charBuffer );

				if ( close == -1 )
					return null;

				if ( close != CHARBUFFSIZE )
					this.eofInPosition = close;

				this.startingPosition = 0;
			}

			String candidate = "";
			int stopIn = CHARBUFFSIZE;

			if ( isLastCharBuff( ) )
			{
				stopIn = this.eofInPosition;
			}

			for ( int i = this.startingPosition; i < stopIn; i++ )
			{
				if ( this.charBuffer[i] == '\n' )
				{
					return readALine( candidate, stopIn, i );
				}
			}

			if ( isLastCharBuff( ) )
			{
				return readLastLine( candidate );
			}

			return readExtraContentOfALine( candidate );
		}

		/**
		 * If a line has not been finished after the current char buffer end is
		 * reach, then read the infinish part of that line from input stream.
		 * 
		 * @param candidate
		 * @return
		 * @throws IOException
		 */
		private String readExtraContentOfALine( String candidate )
				throws IOException
		{
			candidate = candidate
					+ String.copyValueOf( charBuffer,
							this.startingPosition,
							CHARBUFFSIZE - this.startingPosition );

			resetCharBufferStartPosition( );

			String nextLine = readLine( );

			return candidate + ( nextLine == null ? "" : nextLine );
		}

		/**
		 * Read a line from char buffer.
		 * 
		 * @param candidate
		 * @param stopIn
		 * @param i
		 * @return
		 */
		private String readALine( String candidate, int stopIn, int i )
		{
			candidate = candidate
					+ String.copyValueOf( charBuffer, this.startingPosition, i
							- this.startingPosition );
			this.startingPosition = ( i == stopIn - 1 ? -1 : i + 1 );
			return candidate;
		}

		/**
		 * Read the last line of the input stream.
		 * 
		 * @param candidate
		 * @return
		 */
		private String readLastLine( String candidate )
		{
			candidate = candidate
					+ String.copyValueOf( charBuffer,
							this.startingPosition,
							this.eofInPosition - this.startingPosition );
			resetCharBufferStartPosition( );
			return candidate;
		}

		/**
		 * Reset the char buffer start position to -1.
		 * 
		 */
		private void resetCharBufferStartPosition( )
		{
			this.startingPosition = -1;
		}

		/**
		 * Return whether the current char buffer has been finished reading.
		 * 
		 * @return
		 */
		private boolean needRefillCharBuff( )
		{
			return this.startingPosition == -1;
		}

		/**
		 * Return whether there are more char buffer can be read from stream.
		 * 
		 * @return
		 */
		private boolean isLastCharBuff( )
		{
			return this.eofInPosition != -2;
		}

		/**
		 * Close the reader. This method should be call every time the reader is
		 * finish reading.
		 * 
		 * @throws IOException
		 */
		public void close( ) throws IOException
		{
			this.reader.close( );
		}

		/**
		 * New a char buffer with default size.
		 * 
		 * @return
		 */
		private char[] newACharBuff( )
		{
			return new char[CHARBUFFSIZE];
		}
	}
	public void setBoolean(String arg0, boolean arg1) throws OdaException {
		// TODO Auto-generated method stub
		
	}
	public void setBoolean(int arg0, boolean arg1) throws OdaException {
		// TODO Auto-generated method stub
		
	}
	public void setNull(String arg0) throws OdaException {
		// TODO Auto-generated method stub
		
	}
	public void setNull(int arg0) throws OdaException {
		// TODO Auto-generated method stub
		
	}
	@Override
	public void cancel() throws OdaException, UnsupportedOperationException {
		// TODO Auto-generated method stub
		
	}
	@Override
	public String getEffectiveQueryText() {
		// TODO Auto-generated method stub
		return null;
	}
	@Override
	public QuerySpecification getSpecification() {
		// TODO Auto-generated method stub
		return null;
	}
	@Override
	public void setObject(String parameterName, Object value)
			throws OdaException {
		// TODO Auto-generated method stub
		
	}
	@Override
	public void setObject(int parameterId, Object value) throws OdaException {
		// TODO Auto-generated method stub
		
	}
	@Override
	public void setSpecification(QuerySpecification querySpec)
			throws OdaException, UnsupportedOperationException {
		// TODO Auto-generated method stub
		
	}
    
}
